module cde; % CDIFF package main file

algebraic;

% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions are met:
%
%    * Redistributions of source code must retain the relevant copyright
%      notice, this list of conditions and the following disclaimer.
%    * Redistributions in binary form must reproduce the above copyright
%      notice, this list of conditions and the following disclaimer in the
%      documentation and/or other materials provided with the distribution.
%
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
% AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
% THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
% PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNERS OR
% CONTRIBUTORS
% BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
% CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
% SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
% INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
% CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
% ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
% POSSIBILITY OF SUCH DAMAGE.
%

% % *****************************************************************
% Author and maintainer: Raffaele Vitolo
% Dipartimento di Matematica, Universita' del Salento (Lecce, Italy)
% email: raffaele.vitolo@unisalento.it
% web: http://poincare.unisalento.it/vitolo
%
% Version and Date:  v. 1.0, October 2014.
%
% ===============================================================

%
write "This is CDE, a Reduce package for calculus on differential equations";
write "v. 1.0, 10 October 2014";
write "by Raffaele Vitolo";

%
%-----------------------------------------------------------------------------%
% Initialization
%-----------------------------------------------------------------------------%
%
% List of global variables
global '(indep_var dep_var odd_var % Experimental: removed total_order
 principal_der de principal_odd de_odd
 deg_indep_var deg_dep_var deg_odd_var % end of input variables
 n_indep_var all_mind_table all_der_mind all_odd_mind
 all_der_id all_odd_id all_ext
 all_der_oid all_odd_oid repoid repido
 all_der_jid all_odd_jid repjid repidj
 i2m_jetspace i2m_jetspace_odd i2o_jetspace i2o_jetspace_odd
 all_principal_der all_parametric_der all_principal_odd all_parametric_odd
 all_principal_ext all_parametric_ext o2e_jetspace
 repoddext repextodd tot_der id_tot_der eid_tot_der letop
 primary_diffcon_tot primary_diffcon_der primary_diffcon_odd diffcon_der
 diffcon_odd diffcon_comp_der diffcon_comp_odd diffcon_comp_ext
 diffcon_param_der diffcon_param_ext);

% This is the command in the main cdiff file cdiff.red
% create!-package('(cdiff tools21 integrator supervf cdiffx), nil);

% Remember that I should load the assist package using the defautoload mechanism
%
load_package assist;

% This is a temporary solution for loading cdiff and the new files
load_package cdiff;
in "cde_jetspace.red";
in "cde_parametric.red";
in "cde_totalder.red";
in "cde_diffcon.red";
in "cde_tdcalc.red";
in "cde_ansatz.red";

%
%-----------------------------------------------------------------------------%
% Master procedures: creating the environment and saving it.
%-----------------------------------------------------------------------------%
%

procedure cde(jetspace,equation);
   begin;
      % Initialization of global variables
      indep_var:=first jetspace;
      dep_var:=second jetspace;
      odd_var:=third jetspace;
      total_order:=last jetspace;
      principal_der:=first first equation;
      de:=second first equation;
      principal_odd:=first second equation;
      de_odd:=second second equation;
      % Initial tests
      if length(indep_var) equal 0 then rederr
         "Error: missing independent variables";
      if length(dep_var) equal 0 then rederr
         "Error: missing independent variables";
      if length(odd_var) equal 0 then rederr
         "Error: missing odd variables";
      if not fixp total_order then rederr
         "Error: missing maximal order of jetspace";
      if length(principal_der) equal 0 then
         rederr "Error: missing principal derivatives";
      if length(de) equal 0 then rederr
         "Error: missing right-hand side of DE";
      if length(de) neq length(principal_der) then rederr
         "Error: left- and right-hand-side mismatch";
      if length(principal_odd) equal 0 then rederr
         "Error: missing principal odd derivatives" ;
      if length(de_odd) equal 0 then rederr
         "Error: missing right-hand side of odd DE";
      if length(de_odd) neq length(principal_odd) then rederr
         "Error: left- and right-hand-side mismatch";
      % Change next identifiers if you wish to change prefix 
      % to total derivatives
      id_tot_der:=dd;
      eid_tot_der:=td;
      % Building cde data structure:
      write "Building cde data structure in 4 steps:";
      write "1 - Building jetspace ...";
      cde_jetspace(total_order);
      write "2 - Building principal and parametric derivatives ...";
      cde_principal_parametric_derivatives(total_order);
      write "3 - Building total derivatives ...";
      cde_total_derivatives();
      write "4 - Building differential consequences in 3 steps:";
      cde_differential_consequences();
      write "5 - Preparing odd derivatives ...";
      cde_odd_derivatives();
   end;

procedure save_cde_state(statefilename);
   begin scalar n_all_principal_der,n_all_parametric_der,
      n_all_principal_odd,n_all_parametric_odd;
      share state_file;
      if (symbolic not stringp statefilename) then rederr
      	 "Error: save_cde_state parameter must be a string!";
      off echo;
      off nat;
      out <<statefilename>>;
      write "% cde state file";
      write "% even equation:";
      write principal_der,"=",de;
      write "% odd equation";
      write principal_odd,"=",de_odd;
      write " ";
      write "% Settings:";
      write " ";
      write "indep_var:=",indep_var;
      write "dep_var:=",dep_var;
      write "odd_var:=",odd_var;
      write "deg_indep_var:=",deg_indep_var;
      write "deg_dep_var:=",deg_dep_var;
      write "deg_odd_var:=",deg_odd_var;
      write "total_order:=",total_order;
      write "id_tot_der:=",id_tot_der;
      write "principal_der:=",principal_der;
      write "de:=",de;
      write "principal_odd:=",principal_odd;
      write "de_odd:=",de_odd;
      write " ";
      write "% Here starts the output of computations of the program.";
      write " ";
      write "% number of independent variables:";
      write "n_indep_var:=",n_indep_var;
      write "% all multiindexes in table format:";
      write "all_mind_table:=",all_mind_table;
      write "% all even derivatives in multiindex notation:";
      write "all_der_mind:=",all_der_mind;
      write "% all odd derivatives in multiindex notation:";
      write "all_odd_mind:=",all_odd_mind;
      write "% all even derivatives in identifier notation:";
      write "all_der_id:=",all_der_id;
      write "% all odd derivatives in identifier notation:";
      write "all_odd_id:=",all_odd_id;
      write "% all ext symbols:";
      write "all_ext:=",all_ext;
      write "% alist of even derivative coordinates:";
      write "i2m_jetspace:=",i2m_jetspace;
      write "% alist of odd derivative coordinates:";
      write "i2m_jetspace_odd:=",i2m_jetspace_odd;
      write "% alist of even derivative coordinates - order of derivative:";
      write "i2o_jetspace:=",i2o_jetspace;
      write "% alist of odd derivative coordinates - order of derivative:";
      write "i2o_jetspace_odd:=",i2o_jetspace_odd;
      write "% list of even principal derivatives:";
      write "all_principal_der:=",all_principal_der;
      write "% list of even parametric derivatives:";
      write "all_parametric_der:=",all_parametric_der;
      write "% list of odd principal derivatives:";
      write "all_principal_odd:=",all_principal_odd;
      write "% list of odd parametric derivatives:";
      write "all_parametric_odd:=",all_parametric_odd;
      write "% list of ext principal derivatives:";
      write "all_principal_ext:=",all_principal_ext;
      write "% list of ext parametric derivatives:";
      write "all_parametric_ext:=",all_parametric_ext;
      write "% alist of odd derivatives - ext symbols:";
      write "o2e_jetspace:=",o2e_jetspace;
      write "% set of rules to replace odd ders with ext symbols:";
      write "repoddext:=",repoddext;
      write "% set of rules to replace ext symbols with odd ders:";
      write "repextodd:=",repextodd;
      write "% set of identifiers for total derivatives:";
      write "tot_der:=",tot_der;
      write "% set up the auxiliary variable `letop'";
      write "for each el in indep_var do depend letop,el;";
      write "% Characterization of primary differential consequences:";
      write "primary_diffcon_tot:=",primary_diffcon_tot;
      write "primary_diffcon_der:=",primary_diffcon_der;
      write "primary_diffcon_odd:=",primary_diffcon_odd;
      write "% Alists of differential conseq. needed in total derivatives:";
      write "diffcon_der:=",diffcon_der;
      write "diffcon_odd:=",diffcon_odd;
      write "% Principal derivatives computed:";
      write "diffcon_comp_der:=",diffcon_comp_der;
      write "diffcon_comp_odd:=",diffcon_comp_odd;
      write "diffcon_comp_ext:=",diffcon_comp_ext;
      write "% Alists of principal derivatives in terms of parametric expr.:";
      write "diffcon_param_der:=",diffcon_param_der;
      write "diffcon_param_ext:=",diffcon_param_ext;
      write "% Total derivatives declarations:";
      write "for i:=1:n_indep_var do super_vectorfield(part(tot_der,i),append(indep_var,all_der_id),all_ext)";
      write "% Coefficients of total derivatives:";
      n_all_parametric_der:=length(all_parametric_der);
      n_all_principal_der:=length(all_principal_der);
      n_all_parametric_odd:=length(all_parametric_odd);
      n_all_principal_odd:=length(all_principal_odd);
      for i:=1:n_indep_var do
   	 <<
      	    for j:=1:n_indep_var do
      	    <<
	       write "% component of ",part(tot_der,i)," w.r.t. ",
                  part(indep_var,j);
	       write part(tot_der,i),"(",0,",",j,"):=",
   	    	  ev_svf(part(tot_der,i),part(indep_var,j));
      	    >>;
      	    for j:=1:n_all_parametric_der do
      	    <<
	       write "% component of ",part(tot_der,i)," w.r.t. ",
                  part(all_parametric_der,j);
	       write part(tot_der,i),"(",0,",",n_indep_var+j,"):=",
   	    	  ev_svf(part(tot_der,i),part(all_parametric_der,j));
      	    >>;
      	    for j:=1:n_all_principal_der do
      	    <<
	       write "% component of ",part(tot_der,i)," w.r.t. ",
                  part(all_principal_der,j);
	       write part(tot_der,i),
                  "(",0,",",n_indep_var+n_all_parametric_der+j,"):=",
   	    	  ev_svf(part(tot_der,i),part(all_principal_der,j));
      	    >>;
      	    for j:=1:n_all_parametric_odd do
      	    <<
	       write "% component of ",part(tot_der,i)," w.r.t. ",
   	    	  first part(o2e_jetspace,j)," i.e. ",
                  second part(o2e_jetspace,j);
	       write part(tot_der,i),"(",1,",",j,"):=",
   	    	  ev_svf(part(tot_der,i),second part(o2e_jetspace,j));
      	    >>;
            for j:=1+n_all_parametric_odd:
               n_all_parametric_odd+n_all_principal_odd do
      	    <<
	       write "% component of ",part(tot_der,i)," w.r.t. ",
   	    	  first part(o2e_jetspace,j)," i.e. ",
                  second part(o2e_jetspace,j);
	       write part(tot_der,i),"(",1,",",j,"):=",
   	    	  ev_svf(part(tot_der,i),second part(o2e_jetspace,j));
      	    >>;
   	 >>;
   	 write ";end;";
   	 shut <<statefilename>>;
   	 on echo;
   	 on nat;
   end;

endmodule;

end;
