module cde_ansatz; % CDIFF package, computation of the
                    % differential consequences of the initial system of PDEs.

algebraic;

% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions are met:
%
%    * Redistributions of source code must retain the relevant copyright
%      notice, this list of conditions and the following disclaimer.
%    * Redistributions in binary form must reproduce the above copyright
%      notice, this list of conditions and the following disclaimer in the
%      documentation and/or other materials provided with the distribution.
%
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
% AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
% THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
% PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNERS OR
% CONTRIBUTORS
% BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
% CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
% SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
% INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
% CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
% ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
% POSSIBILITY OF SUCH DAMAGE.
%

% % *****************************************************************
% Author and maintainer: Raffaele Vitolo
% Dipartimento di Matematica, Universita' del Salento (Lecce, Italy)
% email: raffaele.vitolo@unisalento.it
% web: http://poincare.unisalento.it/vitolo
%
% Version and Date: v. 1.0, October 2014.
%
% ===============================================================


%
%-----------------------------------------------------------------------------%
% Procedures for construction of ansatz to solve
% determining equations. The aim is to generate monomials
% with specified properties
%-----------------------------------------------------------------------------%

% Run the following procedure in order to initialize
% the grading of jetspace variables.

procedure cde_grading(dindepvar,ddepvar,doddvar);
   begin;
      deg_indep_var:=dindepvar;
      deg_dep_var:=ddepvar;
      deg_odd_var:=doddvar;
      if length(deg_indep_var) neq length(indep_var) then rederr
         "Error: mismatch in degree of independent variables";
      if length(deg_dep_var) neq length(dep_var) then rederr
         "Error: mismatch in degree of dependent variables";
      if length(deg_odd_var) neq length(odd_var) then rederr
         "Error: mismatch in degree of odd dependent variables";
   end;

%
% Section 1: selects a list of variables
%    by which generates monomials up to a given algebraic degree.

% from a set of (even or odd) parametric derivatives `vars' collects
% derivative coordinates of some given dependent variables `depvars'
% and of order equal to `orderofder'
procedure selectvars(par,orderofder,depvars,vars);
   begin scalar fdo;
      fdo:={};
      for each el in vars do
      <<
	 if member(first idtomind(par,el),depvars) then
	 <<
	    if order_of_der(par,el) = orderofder then
               fdo:=el . fdo
	 >>;
      >>;
      return reverse(fdo)
   end;

% The following procedure takes a list of monomials `listmon'
% and collects the product of `var' with all mononials in listmon.
procedure multmon(var,listmon);
   for each el in listmon collect var*el;

% Makes all possible products of first (algebraic) degree variables
% with a list of monomials in the above format.
procedure all_next_mon(fgdo,listmon);
   begin scalar montemp;
      montemp:={};
      for each el in fgdo do
      <<
   	 montemp:=append(multmon(el,listmon),montemp);
      >>;
      return mkset(montemp)
   end;

% makes a list of lists of all monomials of algebraic degree = k <= alg_grad
% from a list of variables `vars'.
procedure mkallmon(alg_grad,vars);
   begin scalar allmon_temp,allmon;
      allmon:=vars;
      allmon_temp:=vars;
      for i:=1:(alg_grad-1) do
      <<
   	 allmon_temp:=all_next_mon(vars,allmon_temp);
   	 allmon:=append(allmon_temp,allmon);
      >>;
      return reverse(allmon)
   end;

% Section 2:  selects a list of graded variables
%    by which generates monomials up to a given scale degree.
% Note that it is more convenient to operate on an alist of graded variables.
% This means using pairs of the form {var,scale_deg}.

% Next procedure finds the scale degree of a given even or odd
% parametric or principal derivative
% using the degree of independent variables
% and the degree of even or odd variables
% given in deg_indep_var, deg_dep_var, deg_odd_var
procedure deg_der(par,der);
   begin scalar deg,der_mind,mind,dep_var_pos,degvar,dvar;
      deg:=0;
      der_mind:=idtomind(par,der);
      mind:=second der_mind;
      if par=0 then
      <<
         degvar:=deg_dep_var;
	 dvar:=dep_var;
      >>
      else
      <<
	 degvar:=deg_odd_var;
	 dvar:=odd_var;
      >>;
      for i:=1:n_indep_var do
      <<
      	 deg:=deg-part(mind,i)*part(deg_indep_var,i);
      >>;
      dep_var_pos:=position(first der_mind,dvar);
      return deg+part(degvar,dep_var_pos)
   end;

% Given a list of parametric or principal derivatives
% next procedure computes their scale degree and arranges an alist
% of the form variable-degree
procedure der2deg(par,vars);
   for each el in vars collect {el,deg_der(par,el)};

% Given an alist of variables and (integer or float) values orders it
% according with increasing (integer or float) value; it uses bubblesort
procedure order_val(vars2value);
   begin scalar l,temp;
      l:=length(vars2value);
      for i:=1:l do
      <<
	 for j:=i+1:l do
	 <<
	    if second(part(vars2value,i))>second(part(vars2value,j)) then
	    <<
	       temp:=part(vars2value,i);
	       vars2value:=part(vars2value,i):=part(vars2value,j);
	       vars2value:=part(vars2value,j):=temp;
	    >>;
	 >>;
      >>;
      return vars2value
   end;

% Within a given alist of graded monomials `allgradmon'
% next procedure finds all possible graded monomials of degree = scale_deg
procedure select_gradmon(gradmon,scale_deg);
   begin scalar scale_deg_gradmon;
      scale_deg_gradmon:={};
      for each el in gradmon do
      <<
      	 if second(el)=scale_deg then
      	    scale_deg_gradmon:=first(el) . scale_deg_gradmon;
      >>;
      return reverse(scale_deg_gradmon)
   end;

% When coordinates have STRICTLY POSITIVE DEGREE,
% the following procedure can be used together with
% graadlijst and mkvarlist1 in order to produce all graded monomials.
% It splits lists of parametric derivatives in sublists according with
% increasing degree starting from degree 1.
procedure der_deg_ordering(par,list_of_der);
   begin scalar list_of_der_deg,alist_of_der_deg,der_i;
      list_of_der_deg:={};
      alist_of_der_deg:=order_val(der2deg(par,list_of_der));
      if second(first(alist_of_der_deg)) <= 0 then
         rederr("Error: parametric derivatives
            do not have strictly positive degree");
      	    for i:=1:second(last(alist_of_der_deg)) do
      	    <<
	       der_i:={};
	       for each el in list_of_der do
	       <<
   	    	  if deg_der(par,el)=i then der_i:=el . der_i
	       >>;
	       der_i:=reverse(der_i);
	       list_of_der_deg:=der_i . list_of_der_deg;
      	    >>;
      	    return reverse(list_of_der_deg)
   end;

% The following procedure yields the graded multiplication of
% two graded monomials.
% Does not work for odd-ext variables.
procedure gradmult(par,firstmon,secondmon);
   {first(firstmon)*first(secondmon),second(firstmon)+second(secondmon)};

% The following procedure is the same as `multmon'
% in the case of alists variable-value,
% where it also takes the sum of the values.
% IMPORTANT: this will not work with odd variables,
% modify it by using super_product for the odd case.
procedure multgradmon(par,gradvar,alistmon);
   for each el in alistmon collect gradmult(par,gradvar,el);

% Makes all possible products of first (algebraic) degree variables
% with an alist of graded monomials.
procedure all_next_gradmon(par,alistvar,alistmon);
   begin scalar gradmontemp,montemp;
      gradmontemp:={};
      montemp:={};
      for each el in alistvar do
      <<
   	 gradmontemp:=
            append(multgradmon(par,el,alistmon),gradmontemp);
      >>;
      return mkset(gradmontemp)
   end;

% Takes a list graded monomials and produces a pair of lists,
% respectively of `good' monomials having the given scale degree
% and of filtered monomials having scale degree less than the given one.
procedure filter_gradmon(par,alistmon,scale_deg);
   begin scalar filtered_gradmon,good_gradmon;
      filtered_gradmon:=order_val(alistmon);
      good_gradmon:={};
      while (last(filtered_gradmon) and
      	 (last(last(filtered_gradmon))>scale_deg)) do
      	    filtered_gradmon:=belast(filtered_gradmon);
      while (last(filtered_gradmon) and
      	 (last(last(filtered_gradmon))=scale_deg)) do
   	 <<
      	    good_gradmon:=last(filtered_gradmon) . good_gradmon;
      	    filtered_gradmon:=belast(filtered_gradmon);
   	 >>;
      return {good_gradmon,filtered_gradmon}
   end;

% Makes an alist of all graded monomials
% of algebraic degree <= alg_deg
% and of scale degree = scale_deg
% from a list of variables `vars'.
% Does not work for odd-ext variables.
% Note that the function mkvarlist1 from Twente CDIFF
% does not work if at least one variable for each scale degree is present.
procedure mkallgradmon(par,vars,alg_deg,scale_deg);
   begin scalar alistvars,gradmon,temp_pair,maxiter,gradmon_temp;
      alistvars:=der2deg(par,vars);
      temp_pair:=filter_gradmon(par,alistvars,scale_deg);
      gradmon:=first temp_pair;
      alistvars:=second temp_pair;
      if alistvars neq {} then
      <<
      	 maxiter:=min(floor(scale_deg/second(first(alistvars))),alg_deg-1);
      	 gradmon_temp:=alistvars;
      	 for i:=1:maxiter do
      	 <<
   	    gradmon_temp:=
               all_next_gradmon(par,alistvars,gradmon_temp);
            temp_pair:=filter_gradmon(par,gradmon_temp,scale_deg);
   	    gradmon:=append(gradmon,first(temp_pair));
            gradmon_temp:=second temp_pair;
      	 >>;
      >>;
      return gradmon
   end;

% Section 3:  using previously formed lists of graded even monomials
%    we generate monomials which are linear in odd variables
%    and have a given scale degree

% Next function takes a list of lists of even monomials `graadmon'
% of degrees 0,1,... and
% a list of parametric odd coordinates `graadlijst_odd'
% of degrees 0,1,...,
% it multiplies all even monomials with odd coordinates of degree >=1
% yielding all products of degree equal to `degree' (>=1).
% It is understood that the first, zero-degree, element of graadlijst_odd
% is always {1}.
procedure mklinodd(graadmon,graadlijst_odd,degree);
   begin scalar linodd,linodd_temp;
      linodd:={};
      for i:=1:degree do
      <<
         for each el in part(graadlijst_odd,i+1) do
         <<
            linodd_temp:={};
            linodd_temp:=for each ell in part(graadmon,degree-i+1)
               collect el*ell;
            linodd:=append(linodd,linodd_temp);
         >>;
      >>;
      return linodd
   end;

% Makes a list of graded monomials, linear in parametric odd coordinates,
% and having degree from degmin to degmax.
procedure mkalllinodd(graadmon,graadlijst_odd,degmin,degmax);
   begin scalar linodd;
      linodd:={};
      for i:=degmin:degmax do
   	 linodd:=append(linodd,mklinodd(graadmon,graadlijst_odd,i));
      return linodd
   end;

% Section 4: same procedures as in S.3 
% but also involving independent variables

procedure mkallgradmon_ind(alg_deg,top_degree);
   % This procedure makes all graded monomials {el,deg} 
   % where el is a monomial of algebraic degree <= alg_deg
   % of independent variables and deg is an integer.
   % It issues an error if graded monomials
   % have degree whose absolute value is higher than top_degree.
   begin scalar allgradmon_temp,allgradmon,indep_gradmon;
      indep_gradmon:=pair(indep_var,deg_indep_var);
      allgradmon:={1,0} . indep_gradmon;
      allgradmon_temp:=indep_gradmon;
      for i:=1:(alg_deg-1) do
      <<
	 allgradmon_temp:=all_next_gradmon(0,indep_gradmon,allgradmon_temp);
         for each el in allgradmon_temp do
         <<
            if abs(second(el))>top_degree then rederr 
               "Error: degree of independent variables lower than top_degree!"
         >>;
	 allgradmon:=append(allgradmon,allgradmon_temp)
      >>;
      return allgradmon
   end;

procedure mklinodd0(graadmon,graadlijst_odd,degree);
   % Next function takes a list of lists of monomials `graadmon'
   % of increasing degrees 0,1,...
   % and a list of parametric odd coordinates `graadlijst_odd'
   % of increasing degrees 0,1,... and multiplies them,
   % yielding all products of degree equal to `degree'.
   begin scalar linodd,linodd_temp;
      linodd:={};
      for i:=0:degree do
      <<
         for each el in part(graadlijst_odd,i+1) do
         <<
            linodd_temp:={};
            linodd_temp:=for each ell in part(graadmon,degree-i+1)
               collect el*ell;
            linodd:=append(linodd,linodd_temp);
         >>;
      >>;
      return linodd
   end;

procedure mklinodd0_ind(grmon_ind,grmon,grlijst_odd,degree);
   % takes graded monomials of independent variables grmon_ind
   % and multiplies them with linear odd monomials of complementary degree
   % in order to produce linear odd monomials possibly with 
   % independent variables and having scale degree `degree'.
   % Note that the monomials can be of zero degree in odd variables; this
   % procedure is suitable for computations of even and odd 
   % conservation laws on coverings at the same time.
   begin scalar tempvar;
      return for each el in grmon_ind join
      <<
	 tempvars:=mklinodd0(grmon,grlijst_odd,degree-second(el));
	 for each ell in tempvars collect first(el)*ell
      >>
   end;

procedure mklinodd_ind(grmon_ind,grmon,grlijst_odd,degree);
   % takes graded monomials of independent variables grmon_ind
   % and multiplies them with linear odd monomials of complementary degree
   % in order to produce linear odd monomials possibly with 
   % independent variables and having scale degree `degree'.
   % Note that the monomials cannot be of zero degree in odd variables;
   % this procedure is suitable for computations of operators.
   begin scalar tempvar;
      return for each el in grmon_ind join
      <<
	 tempvars:=mklinodd(grmon,grlijst_odd,degree-second(el));
	 for each ell in tempvars collect first(el)*ell
      >>
   end;


% Section 5: procedures for solving symmetry equations
% Determines if the order of jetspace is too small for the ansatz; in this
% case total derivative with respect to top order variables yield the
% function 'letop', a dummy function of all variables.
% The presence of such a function in a computation is a signal that
% the computation must be redone in a higher order jetspace.
procedure check_letop(list_expr);
   for each el in list_expr do
   <<
      if not freeof(el,letop) then
         rederr "presence of letop, jetspace too small!"
   >>;

% The following procedures use multi_coeff (from the package `tools').
% It gets all coefficients of monomials appearing in the initial equation(s).
% The coefficients are put into the vector equ after the initial equations.

procedure splitext i;
   begin;
      ll:=operator_coeff(equ i,ext);
      equ(tel:=tel+1):=first ll;
      ll:=rest ll;
      for each el in ll do equ(tel:=tel+1):=second el;
   end;

procedure splitvars i;
   begin;
      ll:=multi_coeff(equ i,vars);
      equ(tel:=tel+1):=first ll;
      ll:=rest ll;
      for each el in ll do equ(tel:=tel+1):=second el;
   end;

% The following procedures prepare tensor components,
% to be used as unknown in crack.

procedure mkid_tensor(tens_id,l_indexes);
% Returns a tensorial identifier tens_id_indexes
begin scalar nind,tempobj;
   tempid:=mkid(tens_id,_);
   while l_indexes neq {} do
   <<
      tempid:=mkid(tempid,first(l_indexes));
      l_indexes:=rest(l_indexes)
   >>;
   return tempid
end;

procedure generate_indexes(num_ind,range_ind);
% Writes a list of all lists of num_ind indexes of range range_ind.
begin scalar templist;
   if num_ind=1 then return for i:=1:range_ind collect {i}
   else
   return for i:=1:range_ind join
      <<
         templist:=generate_indexes(num_ind-1,range_ind);
         for each el in templist collect i . el
      >>
end;

procedure symmetrize_tensor(tens_id,num_ind,range_ind,pos_symm_ind);
% Declares the identifiers of a symmetric tensor to be equal to the identifier
% which has symmetric indexes which are ordered with respect to their value.
begin scalar all_ind,syind,syind_sorted,pair_ind_pos,el_sorted;
   all_ind:=generate_indexes(num_ind,range_ind);
   return for each el in all_ind do
   <<
      syind:=for each pos in pos_symm_ind collect part(el,pos);
      syind_sorted:=sortnumlist(syind);
%      write "syind:=",syind," syind_sorted:=",syind_sorted;
      if syind_sorted neq syind then
      <<
         el_sorted:=el;
         pair_ind_pos:=pair(syind_sorted,pos_symm_ind);
%         write "pair_ind_pos:=",pair_ind_pos;
         for each ind_pos in pair_ind_pos do
            el_sorted:=part(el_sorted,second ind_pos):=first ind_pos;
%         write "el:=",el," el_sorted:=",el_sorted;
         set(mkid_tensor(tens_id,el),mkid_tensor(tens_id,el_sorted))
      >>
   >>
end;

procedure sort_with_sign(numlist);
% Given a list of numbers sorts it
% according with increasing (integer or float) value,
% and attaches to it the sign of the sorting permutation.
   begin scalar l,temp;
      integer sign_perm;
      l:=length(numlist);
      sign_perm:=1;
      for i:=1:l do
	 for j:=i+1:l do
         <<
            if part(numlist,i)=part(numlist,j) then sign_perm:=0;
	    if part(numlist,i)>part(numlist,j) then
	    <<
	       temp:=part(numlist,i);
	       numlist:=part(numlist,i):=part(numlist,j);
	       numlist:=part(numlist,j):=temp;
               sign_perm:=-sign_perm
	    >>
         >>;
      return {sign_perm,numlist}
   end;

procedure skewsymmetrize_tensor(tens_id,num_ind,range_ind,pos_sksymm_ind);
% Declares the identifiers of a symmetric tensor to be equal to the identifier
% which has symmetric indexes which are ordered with respect to their value.
begin scalar all_ind,syind,syind_temp,syind_sorted,syind_sign,
     pair_ind_pos,el_sorted;
   all_ind:=generate_indexes(num_ind,range_ind);
   return for each el in all_ind do
   <<
      syind:=for each pos in pos_sksymm_ind collect part(el,pos);
      syind_temp:=sort_with_sign(syind);
      syind_sorted:=second syind_temp;
      syind_sign:=first syind_temp;
%      write "syind:=",syind," syind_sorted:=",syind_sorted;
      if syind_sign=0 then set(mkid_tensor(tens_id,el),0)
      else if syind_sorted neq syind then
      <<
         el_sorted:=el;
         pair_ind_pos:=pair(syind_sorted,pos_sksymm_ind);
%         write "pair_ind_pos:=",pair_ind_pos;
         for each ind_pos in pair_ind_pos do
            el_sorted:=part(el_sorted,second ind_pos):=first ind_pos;
%         write "el:=",el," el_sorted:=",el_sorted;
         set(mkid_tensor(tens_id,el),syind_sign*mkid_tensor(tens_id,el_sorted))
      >>
   >>
end;

procedure depend_tensor(tens_id,num_ind,range_ind,vars);
% Declares dependency of tensor components
begin scalar all_ind;
   all_ind:=generate_indexes(num_ind,range_ind);
   return for each el in all_ind do
      for each ell in vars do depend(mkid_tensor(tens_id,el),ell)
end;

procedure tensor_components(tens_id,num_ind,range_ind);
% Collects all distinct components of a given tensor
begin scalar all_ind;
   all_ind:=generate_indexes(num_ind,range_ind);
   return for each el in all_ind collect mkid_tensor(tens_id,el)
end;

procedure abs_tensor_components(tens_id,num_ind,range_ind);
% Collects all distinct components of a given tensor UP TO SIGN,
% to be used for skewsymmetric tensors
begin scalar all_ind,tenslist,final_tenslist,ft,rt;
   all_ind:=generate_indexes(num_ind,range_ind);
   tenslist:=for each el in all_ind collect mkid_tensor(tens_id,el);
   tenslist:=delete_all(0,tenslist);
   final_tenslist:={};
   while tenslist neq {} do
   <<
    ft:=first tenslist;
    final_tenslist:=ft . final_tenslist;
    rt:=rest tenslist;
    for each el in rt do
     if ft=el or ft=-el then tenslist:=delete(el,rt)
   >>;
   return reverse final_tenslist
end;

% It would be nice to write the name of the unknown in the file.
% Unfortunately it is always evaluated.
%% procedure write_results(filename,results);
%% begin;
%% off nat;
%% off echo;
%% out <<filename>>;
%% write results:=results;
%% shut <<filename>>;
%% on echo;
%% on nat;
%% return;
%% end;

endmodule;

end;
