module cde_diffcon; % CDIFF package, computation of the
                    % differential consequences of the initial system of PDEs.
algebraic;

% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions are met:
%
%    * Redistributions of source code must retain the relevant copyright
%      notice, this list of conditions and the following disclaimer.
%    * Redistributions in binary form must reproduce the above copyright
%      notice, this list of conditions and the following disclaimer in the
%      documentation and/or other materials provided with the distribution.
%
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
% AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
% THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
% PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNERS OR
% CONTRIBUTORS
% BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
% CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
% SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
% INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
% CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
% ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
% POSSIBILITY OF SUCH DAMAGE.
%

% % *****************************************************************
% Author and maintainer: Raffaele Vitolo
% Dipartimento di Matematica, Universita' del Salento (Lecce, Italy)
% email: raffaele.vitolo@unisalento.it
% web: http://poincare.unisalento.it/vitolo
%
% Version and Date:  v. 1.0, October 2014.
%
% ===============================================================

%-----------------------------------------------------------------------------%
% Computation of differential consequences of E=0
%-----------------------------------------------------------------------------%
%
% We compute differential consequences as total derivatives
% (on the free jetspace!) of de and de_odd,
% then we solve the system of differential consequences
% in terms of parametric derivatives only.
% While solving, other principal derivatives could be needed and computed.

procedure order_diffcon(par,dcon_temp);
   % Given an alist of differential consequences it will be ordered
   % according with decreasing order of principal derivative.
   % It uses bubblesort.
   % The only purpose of this function is to order initial equations
   % according with decreasing order of derivatives.
   begin scalar l,temp;
      l:=length(dcon_temp);
      for i:=1:l do
      <<
	 for j:=i+1:l do
	 <<
	    if order_of_der(par,first(part(dcon_temp,i)))<
               order_of_der(par,first(part(dcon_temp,j))) then
	       <<
	       	  temp:=part(dcon_temp,i);
	       	  dcon_temp:=part(dcon_temp,i):=part(dcon_temp,j);
	       	  dcon_temp:=part(dcon_temp,j):=temp;
	       >>
	 >>
      >>;
      return dcon_temp
   end;

procedure diff_vars(var1,var2);
   % takes the difference of two variables in multiindex notation;
   % returns -1 if the dependent part is different
   % or if the difference of multiindexes contains negative entries;
   % return the difference of the corresponding multiindexes in other cases.
   begin scalar m,n,mmn,mmni;
      if (first(var1) neq first(var2)) then return (-1);
      m:=second var1;
      n:=second var2;
      mmn:={};
      for i:=1:n_indep_var do
      <<
         mmni:=part(m,i)-part(n,i);
         if mmni>=0 then mmn:=mmni . mmn
      >>;
      if length(mmn)<n_indep_var then return (-1)
      else
      	 return reverse mmn
   end;

procedure add_diffcon(par,dcon);
   % Adds the expression of the differential consequences
   % corresponding to the list of principal derivatives `dcon'
   % We assume that diffcon_der and diffcon_odd
   % are made of derivatives in decreasing order
   begin scalar diffcontemp,vartemp,ordtemp,dmind,cnt,pder,exprtemp;
      if par=0 then diffcontemp:=diffcon_der else diffcontemp:=diffcon_odd;
      for each el in dcon do
      <<
         vartemp:=idtomind(par,el);
         ordtemp:=order_of_der(par,el);
         dmind:=-1;
         cnt:=1;
         % Next loop searches the starting point of differential consequence
         % of order less than the order of the variable `el'.
         while order_of_der(par,first(part(diffcontemp,cnt)))>=ordtemp do
            cnt:=cnt+1;
         % Next loop searches for a variable with the same dependent part
         % of the given one and a multiindex which has positive difference
         % with the given one.
         while dmind=-1 do
         <<
            pder:=first(part(diffcontemp,cnt));
            dmind:=diff_vars(vartemp,idtomind(par,pder));
            cnt:=cnt+1
         >>;
         exprtemp:=second(part(diffcontemp,cnt-1));
         for i:=1:n_indep_var do
         <<
            for j:=1:part(dmind,i) do
               exprtemp:=ev_svf(part(tot_der,i),exprtemp)
         >>;
         if not freeof(exprtemp,letop) then exprtemp:=letop;
	 % we insert the differential consequence
         % in such a way to preserve the ordering
         cnt:=1;
         while order_of_der(par,el) <
	    order_of_der(par,first(part(diffcontemp,cnt))) do cnt:=cnt+1;
         diffcontemp:=insert({el,exprtemp},diffcontemp,cnt);
      >>;
      if par=0 then diffcon_der:=diffcontemp
      else diffcon_odd:=diffcontemp;
      if par=0 then
         diffcon_comp_der:=for each el in diffcon_der collect first(el)
      else
         diffcon_comp_odd:=for each el in diffcon_odd collect first(el);
   end;

procedure check_diffcon(dc_der,dc_odd);
   begin scalar dc_extra_der,dc_extra_odd,dc_new_der,dc_new_odd,ell_ext;
      % check if computed differential consequences
      % depend on non-computed (extra) differential consequences.
      % If such non-computed differential consequences exist, compute them.
      dc_extra_der:=diffset(all_principal_der,diffcon_comp_der);
      dc_extra_odd:=diffset(all_principal_odd,diffcon_comp_odd);
      dc_new_der:={};
      dc_new_odd:={};
      for each el in dc_der do
      <<
      	 exprtemp:=second(asfirst(el,diffcon_der));
      	 for each ell in dc_extra_der do
      	 <<
            if not freeof(exprtemp,ell) then dc_new_der:=ell . dc_new_der
      	 >>
      >>;
      for each el in dc_odd do
      <<
      	 exprtemp:=second(asfirst(el,diffcon_odd));
      	 for each ell in dc_extra_der do
      	 <<
            if not freeof(exprtemp,ell) then dc_new_der:=ell . dc_new_der
      	 >>;
      	 for each ell in dc_extra_odd do
      	 <<
            ell_ext:=replace_oddext(ell);
            if not freeof(exprtemp,ell_ext) then
               dc_new_odd:=ell . dc_new_odd
      	 >>
      >>;
      return {mkset(dc_new_der),mkset(dc_new_odd)}
   end;

procedure generate_diffcon();
   % Starting from the primary differential consequences, the code:
   % 1 adds all expressions of differential consequences;
   % 2 makes a scan of them in order to find
   %   secondary differential consequences;
   % 3 if there are secondary diff. cons, it adds all differential consequences
   %   and goes to item 2.
   begin scalar cnt,dependency,dcon_der,dcon_odd,tempder;
      % Initialization of the differential consequences.
      % These are lists of even (odd) principal derivatives
      % paired with their expression.
      % The lists are ordered according with decreasing order of derivatives.
      diffcon_der:=order_diffcon(0,pair(principal_der,de));
      diffcon_odd:=order_diffcon(1,pair(principal_odd,replace_oddext(de_odd)));
      diffcon_comp_der:=for each el in diffcon_der collect first el;
      diffcon_comp_odd:=for each el in diffcon_odd collect first el;
      dependency:=1;
      dcon_der:=primary_diffcon_der;
      dcon_odd:=primary_diffcon_odd;
      while dependency=1 do
      <<
         % compute current differential consequences;
         % it could be that one of the current dcon files is empty.
         if dcon_der neq {} then add_diffcon(0,dcon_der);
         if dcon_odd neq {} then add_diffcon(1,dcon_odd);
         tempder:=check_diffcon(dcon_der,dcon_odd);
         dcon_der:=first(tempder);
         dcon_odd:=second(tempder);
         if (dcon_der={} and dcon_odd={}) then dependency:=0;
      >>;
   end;

procedure replace_diffcon_totder();
   begin scalar tempdcon,var_i,evar_i,expr_i;
      tempdcon:=first primary_diffcon_tot;
      for each el in tempdcon do
      <<
      	 var_i:=last el;
      	 expr_i:= second asfirst(var_i,diffcon_param_der);
      	 set_svf(first el,0,third el,expr_i)
      >>;
      tempdcon:=second primary_diffcon_tot;
      for each el in tempdcon do
      <<
      	 ovar_i:=last el;
      	 evar_i:=second asfirst(ovar_i,o2e_jetspace);
      	 expr_i:= second asfirst(evar_i,diffcon_param_ext);
      	 set_svf(first el,1,third el,expr_i)
      >>;
   end;

procedure cde_differential_consequences();
   begin scalar tempdim,diffcon_comp_der_paramexpr,diffcon_comp_ext_expr,
         diffcon_comp_ext_evenparamexpr,diffcon_comp_ext_paramexpr,
         repprincparam_der,repprincparam_ext;
      % We start by generating differential consequences.
      write "   a - Calculating differential consequences ...";
      generate_diffcon();
      write "   b - Solving the system of differential consequences ...";
      % We solve the system of differential consequences in order
      % to obtain principal derivatives only on the lhs.
      % We do it in steps in order to limit memory occupation.
      % 0 - First of all we prepare new lists of variables
      %     and corresponding expressions:
      diffcon_comp_ext:=for each el in diffcon_comp_odd collect
         replace_oddext(el);
      diffcon_comp_ext_expr:=for each el in diffcon_odd collect
         replace_oddext(second(el));
      %
      % 1 - replaces all computed even principal derivatives in the rhs
      %     in order to obtain a 'clean' system:
      %     even principal derivatives=even parametric expressions;
      repprincparam_der:=for each el in diffcon_der collect
         first el=second el;
      diffcon_comp_der_paramexpr:=(diffcon_comp_der where repprincparam_der);
      diffcon_param_der:=pair(diffcon_comp_der,diffcon_comp_der_paramexpr);
      % renews repprincparam_der to have a 'clean' system:
      repprincparam_der:=for each el in diffcon_param_der collect
         first el=second el;
      %
      % 2 - replaces all even principal derivatives into the prolongation of
      %     odd equations and prepares the substitution list;
      diffcon_comp_ext_evenparamexpr:=
         for each el in diffcon_comp_ext_expr collect
            (el where repprincparam_der);
      clear repprincparam_der;
      %
      % 3 - replaces all computed odd principal derivatives on the rhs
      %     in order to obtain a 'clean' system:
      %     odd principal derivatives=odd parametric expressions;
      tempdim:=length(diffcon_comp_ext);
      repprincparam_ext:=for i:=1:tempdim collect
         part(diffcon_comp_ext,i)=part(diffcon_comp_ext_evenparamexpr,i);
      diffcon_comp_ext_paramexpr:=(diffcon_comp_ext where repprincparam_ext);
      diffcon_param_ext:=pair(diffcon_comp_ext,diffcon_comp_ext_paramexpr);
      clear repprincparam_ext;
      % Finally, we replace the differential consequences
      % in the expression of total derivatives
      write "   c - Replacing differential cons. in total derivatives ...";
      replace_diffcon_totder();
   end;


endmodule;

end;
