module cde_jetspace; % CDIFF package, routines for the definition of jetspace

algebraic;

% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions are met:
%
%    * Redistributions of source code must retain the relevant copyright
%      notice, this list of conditions and the following disclaimer.
%    * Redistributions in binary form must reproduce the above copyright
%      notice, this list of conditions and the following disclaimer in the
%      documentation and/or other materials provided with the distribution.
%
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
% AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
% THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
% PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNERS OR
% CONTRIBUTORS
% BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
% CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
% SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
% INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
% CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
% ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
% POSSIBILITY OF SUCH DAMAGE.
%

% % *****************************************************************
% Author and maintainer: Raffaele Vitolo
% Dipartimento di Matematica, Universita' del Salento (Lecce, Italy)
% email: raffaele.vitolo@unisalento.it
% web: http://poincare.unisalento.it/vitolo
%
% Version and Date:  v. 1.0, October 2014.
%
% ===============================================================


%-----------------------------------------------------------------------------%
% Jetspace coordinates
%-----------------------------------------------------------------------------%

% Here we introduce all jet coordinates up to the order specified
% in the parameter total_order.

% Coordinates are introduced in two formats,
% which have the same name with termination `_mind' or `_id'
% depending on if coordinates are of the form
% {u,{1,2,0}} (multiindex notation)
% or u_x1y2z0 (identifier notation).
% The two formats are necessary for all computations which depend
% on the numeric value of the multiindex.

% We also define odd coordinates, to be used on tangent or cotangent coverings.
% Odd coordinates have the identifier notation and the ext notation,
% to be used in internal computations.

% We also define a number of useful procedures to be used throughout
% the program.

procedure jet_fiber_dim(jorder);
% Compute the number of derivatives of order 'jorder'
% for a jet space with n_ivar indep. vars and n_dvar dep. vars
begin scalar n_ivar,n_dvar;
   n_dvar:=length(dep_var);
   n_ivar:=length(indep_var);
   return n_dvar*(factorial(n_ivar+jorder-1))/
    (factorial(n_ivar)*factorial(jorder-1))
end;

procedure jet_dim(jorder);
% Compute the number of all derivative coordinates up to order 'jorder'
% for a jet space with n_ivar indep. vars and n_dvar dep. vars
begin scalar n_ivar,n_dvar;
   n_dvar:=length(dep_var);
   n_ivar:=length(indep_var);
   n_dvar*(factorial(n_ivar+jorder))/(factorial(n_ivar)*factorial(jorder));
end;

procedure length_multiindex lm;
   % Finds the length of a multiindex
   for each el in lm sum el;

procedure generate_multiindex m;
   % for a given multiindex m generates all multiindexes
   % strictly containing m with length = length m + 1
   % new multiindex appear in the following order:
   % each index is incremented starting from left to right
   % with respect to the order of independent variables.
   begin scalar mip;
      return for i:=1:n_indep_var collect
      <<
	 mip:=m;
	 mip:=part(mip,i):=part(m,i)+1
      >>
   end;

procedure generate_prolongation_table(m,total_order);
   % for a given multiindex m generates a table of the form
   % {mind,list_of_first_prol}
   % where |m| <= total_order
   begin scalar temp_m_list,temp_table_rows,all_table_rows;
      temp_m_list:={m};
      all_table_rows:={};
      while length_multiindex(first(temp_m_list))<total_order do
      <<
	 temp_table_rows:={};
	 temp_table_rows:=for each el in temp_m_list collect
	    {el,generate_multiindex(el)};
	 temp_m_list:=mkset(for each ell in temp_table_rows join second ell);
	 all_table_rows:=append(all_table_rows,temp_table_rows);
      >>;
      temp_m_list:=for each elll in temp_m_list collect {elll,{}};
      return append(all_table_rows,temp_m_list)
   end;

procedure join_dvar_mind(dvar,tm);
   % from a list of dependent variables and a table of multiindexes
   % produces a list with all derivative coordinates in multiindex notation
   % the list is ordered according with the length of multiindexes and dvar
   for each el in tm join
   <<
      for each ell in dvar collect {ell,first(el)}
   >>;

procedure mind_to_oid der;
   % converts a derivative coordinate from the multiindex notation
   % to the identifier notation - ex.: u_t1x2y0
   begin scalar mip;
      mip:=for i:=1:n_indep_var join {part(indep_var,i),part(second der,i)};
      return list_to_ids(first der . (_ . mip))
   end;

procedure mind_to_eid der;
   % converts a derivative coordinate from the multiindex notation
   % to the external identifier notation - ex.: u_t2x
   begin scalar mip,tempind;
      mip:=for i:=1:n_indep_var join
      <<
         tempind:=part(second der,i);
         if tempind=0 then {}
            else if tempind=1 then {part(indep_var,i)}
            else {tempind,part(indep_var,i)}
      >>;
      return
         if mip={} then first der else list_to_ids(first der . (_ . mip))
   end;

procedure mind_to_jid der;
   % converts a derivative coordinate from the multiindex notation
   % to the external identifier notation by Jets - u_txx
   begin scalar mip,tempind,tempvar;
      mip:=for i:=1:n_indep_var join
      <<
         tempind:=part(second der,i);
         tempvar:=part(indep_var,i);
         if tempind=0 then {}
            else for i:=1:tempind collect tempvar
      >>;
      return
         if mip={} then first der else list_to_ids(first der . (_ . mip))
   end;

procedure idtomind(par,der_id);
   % given a derivative coordinate in identifier notation, finds the
   % corresponding multiindex notation. Works for even (par=0) and odd (par=1)
   % jetspace coordinates.
   begin scalar der_mind;
      if par=0 then
         der_mind:=second asfirst(der_id,i2m_jetspace)
      else
         der_mind:=second asfirst(der_id,i2m_jetspace_odd);
      return der_mind
   end;

procedure order_of_der_mind(der_mind);
   % finds the order of a given (odd) jetspace coordinate
   % in multiindex notation
   length_multiindex(second der_mind);

procedure order_of_der(par,der_id);
   % finds the order of a given (odd) jetspace coordinate
   begin scalar ord_der;
      if par=0 then
      	 ord_der:=second asfirst(der_id,i2o_jetspace)
      else
      	 ord_der:=second asfirst(der_id,i2o_jetspace_odd);
      return ord_der
   end;

procedure cde_jetspace(total_order);
   begin scalar n_all_odd_id,tempeidder,tempeidodd;
      n_indep_var:=length(indep_var);
      if n_indep_var equal 0 then rederr "Error: missing dependent variables";
      % produces a list of all multiindexes obtained from the zero multiindex
      % and fulfilling the condition length<=total_order
      % then generates all derivative coordinates in multiindex notation
      all_mind_table:=
         generate_prolongation_table(mklist(length(indep_var)),total_order);
      all_der_mind:=join_dvar_mind(dep_var,all_mind_table);
      all_odd_mind:=join_dvar_mind(odd_var,all_mind_table);
      % produces lists with all derivative coordinates
      % of dependent and odd coordinates in identifier notation
      all_der_id:=map(mind_to_eid,all_der_mind);
      all_odd_id:=map(mind_to_eid,all_odd_mind);
      % generates a list of all ext variables;
      n_all_odd_id:=length(all_odd_id);
      all_ext:=for i:=1:n_all_odd_id collect ext i;
      % makes association lists which are used for conversion
      % from identifiers to multiindexes.
      i2m_jetspace:=pair(all_der_id,all_der_mind);
      i2m_jetspace_odd:=pair(all_odd_id,all_odd_mind);
      % makes association lists which are used
      % for the computation of the order of a derivative coordinate
      i2o_jetspace:=pair(all_der_id,map(order_of_der_mind,all_der_mind));
      i2o_jetspace_odd:=pair(all_odd_id,map(order_of_der_mind,all_odd_mind));
   end;

procedure cde_old_jetspace(total_order);
% Prepares 'old-style' coordinates for replacement with new ones
% and vice-versa.
   begin scalar n_all_odd_id,tempeidder,tempeidodd;
      n_all_odd_id:=length(all_odd_id);
      all_der_oid:=map(mind_to_oid,all_der_mind);
      all_odd_oid:=map(mind_to_oid,all_odd_mind);
      % makes lists of relations which are used for conversion
      % from external to internal identifiers
      tempoidder:=for i:=1:n_all_odd_id collect
         part(all_der_oid,i)=part(all_der_id,i);
      tempoidodd:=for i:=1:n_all_odd_id collect
         part(all_odd_oid,i)=part(all_odd_id,i);
      repoid:=append(tempoidder,tempoidodd);
      tempoidder:=for i:=1:n_all_odd_id collect
         part(all_der_id,i)=part(all_der_oid,i);
      tempoidodd:=for i:=1:n_all_odd_id collect
         part(all_odd_id,i)=part(all_odd_oid,i);
      repido:=append(tempoidder,tempoidodd);
   end;

procedure jets_jetspace(total_order);
% Prepares Jets-style coordinates for replacement with CDIFF coordinates
% and vice-versa.
   begin scalar n_all_odd_id,tempjidder,tempjidodd;
      n_all_odd_id:=length(all_odd_id);
      all_der_jid:=map(mind_to_jid,all_der_mind);
      all_odd_jid:=map(mind_to_jid,all_odd_mind);
      % makes lists of relations which are used for conversion
      % from external to internal identifiers
      tempjidder:=for i:=1:n_all_odd_id collect
         part(all_der_jid,i)=part(all_der_id,i);
      tempjidodd:=for i:=1:n_all_odd_id collect
         part(all_odd_jid,i)=part(all_odd_id,i);
      repjid:=append(tempjidder,tempjidodd);
      tempjidder:=for i:=1:n_all_odd_id collect
         part(all_der_id,i)=part(all_der_jid,i);
      tempjidodd:=for i:=1:n_all_odd_id collect
         part(all_odd_id,i)=part(all_odd_jid,i);
      repidj:=append(tempjidder,tempjidodd);
   end;

procedure replace_old_id listexpr;
   sub(repoid,listexpr);
procedure replace_id_old listexpr;
   sub(repido,listexpr);

procedure replace_jet_id listexpr;
   sub(repjid,listexpr);
procedure replace_id_jet listexpr;
   sub(repidj,listexpr);

endmodule;

end;