module cde_parametric; % CDIFF package, routines for the computation
                     % of principal and parametric derivatives.
algebraic;

% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions are met:
%
%    * Redistributions of source code must retain the relevant copyright
%      notice, this list of conditions and the following disclaimer.
%    * Redistributions in binary form must reproduce the above copyright
%      notice, this list of conditions and the following disclaimer in the
%      documentation and/or other materials provided with the distribution.
%
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
% AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
% THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
% PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNERS OR
% CONTRIBUTORS
% BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
% CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
% SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
% INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
% CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
% ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
% POSSIBILITY OF SUCH DAMAGE.
%

% % *****************************************************************
% Author and maintainer: Raffaele Vitolo
% Dipartimento di Matematica, Universita' del Salento (Lecce, Italy)
% email: raffaele.vitolo@unisalento.it
% web: http://poincare.unisalento.it/vitolo
%
% Version and Date:  v. 1.0, October 2014.
%
% ===============================================================

%-----------------------------------------------------------------------------%
% Computation of all parametric (and principal) derivatives (even and odd)
%-----------------------------------------------------------------------------%
procedure process_principal_der(par,l_var);
   % It is used to generate all principal derivatives.
   % It takes a list of derivative coordinates and returns a list of lists,
   % each sublist is a list of derivatives of the same dependent coordinate
   % in multiindex notation. Since it is used on initial principal derivatives
   % (lhs of given equations) there is no need to optimize it.
   begin scalar l_mind,dep_var_temp,l_temp,l_var_processed;
      dep_var_temp:={};
      l_var_processed:={};
      l_mind:=for each el in l_var collect idtomind(par,el);
      if par=0 then dep_var_temp:=dep_var else dep_var_temp:=odd_var;
      for each el in dep_var_temp do
      <<
      	 l_temp:={};
      	 for each ell in l_mind do
      	 <<
            if el=first ell then l_temp:=ell . l_temp
      	 >>;
      	 if l_temp neq {} then
         <<
            l_temp:=reverse(l_temp);
            l_var_processed:=l_temp . l_var_processed
         >>
      >>;
      return reverse(l_var_processed)
   end;

procedure generate_principal_subtable(par,list_pd,total_order);
   % Given a list of principal derivatives of a single dependent variable
   % in multiindex notation, it generates all differential consequences
   % (i.e. the output of process_principal_der)
   % in the form of a pair of the given dep. var.
   % and a subtable of all_mind_table
   begin scalar dep_var_temp,principal_mtable_temp,principal_mtable_ord,ord_el;
      dep_var_temp:=first(first(list_pd));
      principal_mtable_temp:=mkset(
         for each el in list_pd join
            generate_prolongation_table(second(el),total_order)
            );
      principal_mtable_ord:=for i:=0:total_order collect {};
      for each el in principal_mtable_temp do
      <<
         ord_el:=length_multiindex(first el);
         principal_mtable_ord:=part(principal_mtable_ord,ord_el+1)
            :=el . part(principal_mtable_ord,ord_el+1)
      >>;
      return {dep_var_temp,principal_mtable_ord}
   end;

procedure generate_all_principal_subtables(par,prin_der,total_order);
   % Makes a subtable for each dependent variable for which
   % a principal derivative is given
   begin scalar principal_der_split;
      principal_der_split:=process_principal_der(par,prin_der);
      return for each el in principal_der_split collect
      <<
         generate_principal_subtable(par,el,total_order)
      >>
   end;

procedure generate_all_principal_der(par,split_subtable,total_order);
   % Generates principal derivatives from subtables
   begin scalar dep_var_temp,principal_mtable_temp,principal_der_temp,
         i2m_temp,apd,apd_tot;
      apd:={};
      if par=0 then i2m_temp:=i2m_jetspace else i2m_temp:=i2m_jetspace_odd;
      for each el in split_subtable do
      <<
         dep_var_temp:=first el;
         principal_mtable_temp:=second el;
         principal_der_temp:=for i:=0:total_order collect {};
         for i:=0:total_order do
         <<
            if (part(principal_mtable_temp,i+1) neq {}) then
            <<
               principal_der_temp:=part(principal_der_temp,i+1):=
               	  (for each ell in part(principal_mtable_temp,i+1) collect
                     first aslast({dep_var_temp,first ell},i2m_temp))
            >>
         >>;
         apd:=principal_der_temp .apd
      >>;
      apd:=reverse apd;
      apd_tot:=for i:=0:total_order collect {};
      for i:=0:total_order do
      <<
         apd_tot:=part(apd_tot,i+1):=(for each el in apd join part(el,i+1))
      >>;
      return apd_tot
   end;

procedure cde_principal_parametric_derivatives(total_order);
begin scalar all_principal_der_subtables,all_principal_der_split,
      all_principal_odd_subtables,all_principal_odd_split,
      n_all_parametric_odd,n_all_principal_odd,tempparam,tempprinc;
   % Even variables:
   all_principal_der_subtables:=
      generate_all_principal_subtables(0,principal_der,total_order);
   all_principal_der_split:=
      generate_all_principal_der(0,all_principal_der_subtables,total_order);
   all_principal_der:=for each el in all_principal_der_split join el;
   all_parametric_der:=diffset(all_der_id,all_principal_der);
   % Odd variables:
   all_principal_odd_subtables:=
      generate_all_principal_subtables(1,principal_odd,total_order);
   all_principal_odd_split:=
      generate_all_principal_der(1,all_principal_odd_subtables,total_order);
   all_principal_odd:=for each el in all_principal_odd_split join el;
   all_parametric_odd:=diffset(all_odd_id,all_principal_odd);
   % Ext variables:
   n_all_parametric_odd:=length(all_parametric_odd);
   n_all_principal_odd:=length(all_principal_odd);
   all_parametric_ext:=for i:=1:n_all_parametric_odd collect ext i;
   all_principal_ext:=
      for i:=1:n_all_principal_odd collect ext(i+n_all_parametric_odd);
   % Substitution odd->ext and ext->odd, on jetspace:
   tempparam:=pair(all_parametric_odd,all_parametric_ext);
   tempprinc:=pair(all_principal_odd,all_principal_ext);
   o2e_jetspace:=append(tempparam,tempprinc);
   tempparam:=for i:=1:n_all_parametric_odd collect
      part(all_parametric_odd,i)=part(all_parametric_ext,i);
   tempprinc:=for i:=1:n_all_principal_odd
      collect part(all_principal_odd,i)=part(all_principal_ext,i);
   repoddext:=append(tempparam,tempprinc);
   tempparam:=for i:=1:n_all_parametric_odd collect
      part(all_parametric_ext,i)=part(all_parametric_odd,i);
   tempprinc:=for i:=1:n_all_principal_odd
      collect part(all_principal_ext,i)=part(all_principal_odd,i);
   repextodd:=append(tempparam,tempprinc);
end;

procedure replace_oddext listexpr;
   sub(repoddext,listexpr);
procedure replace_extodd listexpr;
   sub(repextodd,listexpr);

endmodule;

end;