module cde_tdcalc; % CDIFF package, calculus with total derivatives

algebraic;

% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions are met:
%
%    * Redistributions of source code must retain the relevant copyright
%      notice, this list of conditions and the following disclaimer.
%    * Redistributions in binary form must reproduce the above copyright
%      notice, this list of conditions and the following disclaimer in the
%      documentation and/or other materials provided with the distribution.
%
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
% AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
% THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
% PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNERS OR
% CONTRIBUTORS
% BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
% CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
% SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
% INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
% CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
% ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
% POSSIBILITY OF SUCH DAMAGE.
%

% % *****************************************************************
% Author and maintainer: Raffaele Vitolo
% Dipartimento di Matematica, Universita' del Salento (Lecce, Italy)
% email: raffaele.vitolo@unisalento.it
% web: http://poincare.unisalento.it/vitolo
%
% Version and Date:  v. 1.0, October 2014.
%
% ===============================================================

%
%-----------------------------------------------------------------------------%
% Calculus with total derivatives
%-----------------------------------------------------------------------------%
%

% Remember: add to global variables svf_ext

procedure cde_odd_derivatives();
% Definition of the derivatives with respect to odd variables.
begin scalar n_ext,delta_ij;
   delta_ij:=0;
   n_ext:=length(all_ext);
   for i:=1:n_ext do super_vectorfield(mkid(svf_ext,i),{},all_ext);
   for i:=1:n_ext do
    for j:=1:n_ext do
    <<
       if i=j then delta_ij:=1 else delta_ij:=0;
       set_svf(mkid(svf_ext,i),1,j,delta_ij)
    >>;
end;

procedure df_odd(expr_odd,oddvar);
   % Definition of the derivative operator with respect to odd variables.
   % Everything is converted in the internal ext notation,
   % then back to external.
   begin scalar pos_oddvar;
      pos_oddvar:=position(replace_oddext(oddvar),all_ext);
      return replace_extodd(
       ev_svf(mkid(svf_ext,pos_oddvar),replace_oddext(expr_odd))
      )
   end;

procedure df_ext(expr_ext,extvar);
   % Definition of the derivative operator with respect to odd variables.
   % Everything is in the internal ext notation, should be faster than df_odd
   begin scalar pos_extvar;
      pos_extvar:=position(extvar,all_ext);
      return ev_svf(mkid(svf_ext,pos_extvar),expr_ext)
   end;

procedure pvar_df_old(par,density,component);
% Definition of the partial Euler-Lagrange operator.
% It is of the form (-1)^abs(sigma)*D_sigma df(,u^i_sigma)
% where u^i=component and sigma is such that u^i_sigma is parametric.
% When the equation is evolutionary the component of sigma
% corresponding to the evolutionary variable is zero.
% In this case the partial Euler-Lagrange operator yields
% generating functions of conservation laws.
% When the equation is empty we have the full Euler-Lagrange operator.
% It will work only on linear functions of odd variables;
% in order to make it work with all functions of odd variables
% the derivative wrt one odd variable has to be redefined,
% possibly using super_vectorfield.
begin scalar tempvar,tempmind,exprtemp,result;
   result:=0;
   if par=0 then
   <<
      for each el in all_parametric_der do
      <<
         tempvar:=idtomind(par,el);
         if first(tempvar)=component then
         <<
            tempmind:=second tempvar;
            exprtemp:=(-1)**(length_multiindex(tempmind))*df(density,el);
            for i:=1:n_indep_var do
            <<
               for j:=1:part(tempmind,i) do
               exprtemp:=ev_svf(part(tot_der,i),exprtemp);
            >>;
            result:=result+exprtemp
         >>
      >>
   >>
   else
   <<
      for each el in all_parametric_ext sum
     <<
        tempvar:=idtomind(par,replace_extodd(el));
        if first(tempvar)=component then
        <<
           tempmind:=second tempvar;
           exprtemp:=(-1)**(length_multiindex(tempmind))*df(density,el);
           for i:=1:n_indep_var do
           <<
              for j:=1:part(tempmind,i) do
               exprtemp:=ev_svf(part(tot_der,i),exprtemp)
           >>;
           result:=result+exprtemp;
        >>
     >>
   >>;
   return result
end;

procedure pvar_df_notoptimized(par,density_odd,component);
% Definition of the partial Euler-Lagrange operator.
% It is of the form (-1)^abs(sigma)*D_sigma df(,u^i_sigma)
% where u^i=component and sigma is such that u^i_sigma is parametric.
% When the equation is evolutionary the component of sigma
% corresponding to the evolutionary variable is zero.
% In this case the partial Euler-Lagrange operator yields
% generating functions of conservation laws.
% When the equation is empty we have the full Euler-Lagrange operator.
% NOTE: the input density density_odd and the component
% must be in odd variables!
begin scalar tempvar,tempmind,exprtemp,density_ext,result;
   result:=0;
   if par=0 then
   <<
      for each el in all_parametric_der do
      <<
         tempvar:=idtomind(par,el);
         if first(tempvar)=component then
         <<
            tempmind:=second tempvar;
            exprtemp:=(-1)**(length_multiindex(tempmind))*df(density_odd,el);
            for i:=1:n_indep_var do
            <<
               for j:=1:part(tempmind,i) do
               exprtemp:=ev_svf(part(tot_der,i),exprtemp);
            >>;
            result:=result+exprtemp
         >>
      >>
   >>
   else
   <<
      density_ext:=replace_oddext(density_odd);
      for each el in all_parametric_ext sum
     <<
        tempvar:=idtomind(par,replace_extodd(el));
        if first(tempvar)=component then
        <<
           tempmind:=second tempvar;
           exprtemp:=(-1)**(length_multiindex(tempmind))*df_ext(density_ext,el);
           for i:=1:n_indep_var do
           <<
              for j:=1:part(tempmind,i) do
               exprtemp:=ev_svf(part(tot_der,i),exprtemp)
           >>;
           result:=result+exprtemp;
        >>
     >>
   >>;
   return replace_extodd(result)
end;


procedure pvar_df(par,density_odd,component);
% Definition of the partial Euler-Lagrange operator.
% It is of the form (-1)^abs(sigma)*D_sigma df(,u^i_sigma)
% where u^i=component and sigma is such that u^i_sigma is parametric.
% When the equation is evolutionary the component of sigma
% corresponding to the evolutionary variable is zero.
% In this case the partial Euler-Lagrange operator yields
% generating functions of conservation laws.
% When the equation is empty we have the full Euler-Lagrange operator.
% NOTE: the input density density_odd and the component
% must be in odd variables!
begin scalar tempvar,tempmind,exprtemp,density_ext,result;
   result:={};
   if par=0 then
   <<
     density_ext:=replace_oddext(density_odd);
      for each el in all_parametric_der do
      <<
         tempvar:=idtomind(par,el);
         if first(tempvar)=component then
         <<
            tempmind:=second tempvar;
            exprtemp:=(-1)**(length_multiindex(tempmind))*df(density_ext,el);
            for i:=1:n_indep_var do
            <<
               for j:=1:part(tempmind,i) do
               exprtemp:=ev_svf(part(tot_der,i),exprtemp);
            >>;
            result:=exprtemp . result
         >>
      >>
   >>
   else
   <<
      density_ext:=replace_oddext(density_odd);
      for each el in all_parametric_ext do
     <<
        tempvar:=idtomind(par,replace_extodd(el));
        if first(tempvar)=component then
        <<
           tempmind:=second tempvar;
           exprtemp:=(-1)**(length_multiindex(tempmind))*df_ext(density_ext,el);
           for i:=1:n_indep_var do
           <<
              for j:=1:part(tempmind,i) do
               exprtemp:=ev_svf(part(tot_der,i),exprtemp)
           >>;
           result:=exprtemp . result
        >>
     >>
   >>;
   result:=part(result,0):=plus;
   return replace_extodd(result)
end;

procedure euler_df(density);
% Computes the Euler operator of a (possibly odd) density
% The output is in the form of a list of two sublists,
% the first contains variational derivatives with respect to even variables,
% the second contains variational derivatives with respect to odd variables.
begin scalar even_vder,odd_vder;
  even_vder:=
   for i:=1:length(dep_var) collect pvar_df(0,density,part(dep_var,i));
  odd_vder:=
     for i:=1:length(odd_var) collect pvar_df(1,density,part(odd_var,i));
  return {even_vder,odd_vder}
end;

procedure conv_cdiff2genfun(cdiff_op);
% Convert a C-differential operator into the corresponding generating function
% The operator must be input as a matrix of scalar C-differential operators.
% It is assumed that the column index has a range equal to the number
% of odd variables.
% TEMPORARY: I ASSUME THAT THE MATRIX IS SQUARE!!!!
begin scalar n_odd_var;
  n_odd_var:=length(odd_var);
  return
   for i:=1:n_odd_var collect
   for j:=1:n_odd_var sum
   replace_extodd(cdiff_op(i,j,replace_oddext(part(odd_var,j))))
end;

procedure conv_genfun2biv(genfun_odd);
% Convert a generating function of an operator (in odd variables), i.e.
% a linear function on odd variables, into a bivector.
% The generating function must be input as a list.
begin scalar n_comp_genfun,genfun_ext;
  n_comp_genfun:=length(genfun_odd);
  genfun_ext:=replace_oddext(genfun_odd);
  if n_comp_genfun neq length(odd_var) then rederr
   "Error: The number of components is not equal to the number of odd vars.";
  return for i:=1:n_comp_genfun sum
   super_product(part(genfun_ext,i),replace_oddext(part(odd_var,i)))
end;

procedure ev_operator(el,ell);
   % This procedure is needed because REDUCE refuses
   % to define or evaluate an operator in an iteration by a direct call.
   el(ell);

procedure conv_oper2genfun(oper_components);
% Convert a C-differential operator,
% with values in sections of the odd variable space,
% in the form of a list of operators,
% one for each component, into a a generating function,
% i.e. a linear function in odd variables.
begin scalar n_oper_components;
  n_oper_components:=length(oper_components);
  if n_oper_components neq length(dep_var) then rederr
    "Error: The number of operators is not equal to the number of odd vars.";
  return for i:=1:n_oper_components collect
    ev_operator(part(oper_components,i),replace_oddext(part(odd_var,i)))
end;

% procedure conv_genfun2oper(genfun);
% Convert a generating function, i.e., a linear function in odd variables,
% into a C-differential operator.

procedure check_skewadjoint(biv);
% Check if a given bivector is skewadjoint using the formula from KKV, page 284
begin scalar res;
  res:=replace_extodd(
   for each el in odd_var sum
    super_product(
     replace_oddext(pvar_df(1,biv,el)),replace_oddext(el)
    )
  );
  return res+2*biv;
end;

procedure schouten_bracket(biv1,biv2);
% Compute the Schouten bracket of two bivectors; bivectors are input
% in ext notation (for the moment ...)
% See Kersten, Krasilshchik and Verbovetsky for formulae
begin
  n_dep_var:=length(dep_var);
  if n_dep_var neq length(odd_var) then rederr
    "Error: number of even and odd variables mismatch!";
  tempterm1:=for i:=1:n_dep_var collect
   super_product(
    replace_oddext(pvar_df(0,biv1,part(dep_var,i))),
    replace_oddext(pvar_df(1,biv2,part(odd_var,i)))
   );
  tempterm2:=for i:=1:n_dep_var collect
   super_product(
    replace_oddext(pvar_df(0,biv2,part(dep_var,i))),
    replace_oddext(pvar_df(1,biv1,part(odd_var,i)))
   );
  tempterm1:=part(tempterm1,0):=plus;
  tempterm2:=part(tempterm2,0):=plus;
  return euler_df(tempterm1+tempterm2)
end;

endmodule;

end;
